theory Metatheory
imports Typing Semantics

begin

lemma weakening:
  assumes "\<Gamma> \<turnstile> e : \<tau>"
  and "\<Gamma> \<union> \<Gamma>' \<Turnstile> OK"
  shows "\<Gamma> \<union> \<Gamma>' \<turnstile> e : \<tau>"
using prems by (induct arbitrary: \<Gamma>', auto simp add: wf_env_cons dom_def)

lemma substitution:
  fixes n :: nat
  assumes "size e = n"
  and "\<Gamma> \<turnstile> e' : \<tau>'"
  and "\<Gamma>,(x:\<tau>') \<turnstile> vsubst_tm (TmFreeVar x) i e : \<tau>"
  and "\<not> x free in e"
  shows "\<Gamma> \<turnstile> vsubst_tm e' i e : \<tau>"
using prems proof (induct n arbitrary: e \<tau> i \<Gamma> M rule: nat_less_induct)
  case (1 n_y e_y \<tau>)
  have Hgok: "\<Gamma> \<Turnstile> OK" using prems by (auto simp add: ty_wf_env)
  show ?case proof (cases e_y)
    case TmUnit thus ?thesis using Hgok prems by auto
  next
    case TmFreeVar thus ?thesis using Hgok prems by auto
  next
    case (TmVar j)
    show ?thesis proof (cases "j=i")
      case True
      hence Ht: "\<Gamma>,(x:\<tau>') \<turnstile> TmFreeVar x : \<tau>" using prems by auto
      hence "\<tau> = \<tau>'" using wf_env_def[of "\<Gamma>,(x:\<tau>')"] by fast
      thus ?thesis using True TmVar by auto
    next
      case False
      hence "\<Gamma>,(x:\<tau>') \<turnstile> TmVar j : \<tau>" using prems by auto
      thus ?thesis by (cases, auto)
    qed
  next
    case (TmAbs \<tau>_y e_b)
    let ?e_bx = "vsubst_tm (TmFreeVar x) (Suc i) e_b"
    let ?e_b' = "vsubst_tm e' (Suc i) e_b"
    have Ht: "\<Gamma>,(x:\<tau>') \<turnstile> \<lambda>[\<tau>_y]. ?e_bx : \<tau>" using prems by auto
    then obtain \<tau>_b L where Heq: "\<tau> = \<tau>_y \<rightarrow> \<tau>_b"
      and Hfin: "finite L"
      and Htyp: "\<forall>z. \<not> z free in ?e_bx \<and> z \<notin> L \<and> \<not> \<Gamma>,(x:\<tau>') defines z \<longrightarrow>
                       \<Gamma>,(x:\<tau>'),(z:\<tau>_y) \<turnstile> freshen_tm z ?e_bx : \<tau>_b" by auto
    let ?L = "L \<union> {x} \<union> set (free_tm ?e_bx)"
    have Hfin': "finite ?L" using Hfin by auto
    have Htyp': "\<forall>z. \<not> z free in ?e_b' \<and> z \<notin> ?L \<and> \<not> \<Gamma> defines z \<longrightarrow>
                   \<Gamma>,(z:\<tau>_y) \<turnstile> freshen_tm z ?e_b' : \<tau>_b" proof auto
      fix z
      assume Hnzf: "\<not> z free in ?e_b'"
        and Hnzl: "z \<notin> L"
        and Hnzbx: "\<not> z free in ?e_bx"
        and Hnzx: "z \<noteq> x"
        and Hnzg: "\<not> \<Gamma> defines z"
      have A1: "size (freshen_tm z e_b) < n_y" using prems by auto
      have "\<Gamma>,(z:\<tau>_y) \<Turnstile> OK" using Hgok Hnzg by (auto simp add: wf_env_def dom_def)
      hence A2: "\<Gamma>,(z:\<tau>_y) \<turnstile> e' : \<tau>'" using prems weakening[of _ _ _ "{(z\<rhd>\<tau>_y)}"] by auto
      have "\<not> z free in ?e_bx" and "z \<notin> L" and "\<not> \<Gamma>,(x:\<tau>') defines z"
        using Hnzbx Hnzl Hnzg Hnzx by (auto simp add: dom_def)
      hence "\<Gamma>,(x:\<tau>'),(z:\<tau>_y) \<turnstile> freshen_tm z ?e_bx : \<tau>_b" using Htyp by auto
      hence A3: "\<Gamma>,(z:\<tau>_y),(x:\<tau>') \<turnstile> vsubst_tm (TmFreeVar x) (Suc i) (freshen_tm z e_b) : \<tau>_b"
        by (auto simp add: env_permute subst_subst_swap[of "0" "Suc i"])
      have A4: "\<not> x free in (freshen_tm z e_b)" using prems by (auto simp add: not_free_subst)
      have "\<Gamma>,(z:\<tau>_y) \<turnstile> vsubst_tm e' (Suc i) (freshen_tm z e_b) : \<tau>_b"
        using "1.hyps" A1 A2 A3 A4 by auto
      moreover
      have "\<forall>k. vsubst_tm (TmFreeVar z) k e' = e'" using prems by (auto simp add: ty_same_subst)
      ultimately
      show "\<Gamma>,(z:\<tau>_y) \<turnstile> vsubst_tm (TmFreeVar z) 0 ?e_b' : \<tau>_b" by (auto simp add: subst_subst_swap)
    qed
    have "\<Gamma> \<turnstile> \<lambda>[\<tau>_y]. ?e_b' : \<tau>_y \<rightarrow> \<tau>_b" using Hgok Hfin' Htyp' ..
    thus ?thesis using TmAbs Heq by auto
  next
    case (TmApp e1 e2)
    have Hsize1: "size e1 < n_y" and Hsize2: "size e2 < n_y" using "1.prems" TmApp by auto
    have Hfree1: "\<not> x free in e1" and Hfree2: "\<not> x free in e2" using "1.prems" TmApp by auto
    let ?e1x = "vsubst_tm (TmFreeVar x) i e1"
    let ?e2x = "vsubst_tm (TmFreeVar x) i e2"
    have "\<Gamma>,(x:\<tau>') \<turnstile> ?e1x \<cdot> ?e2x : \<tau>" using prems by auto
    then obtain \<tau>1 where Ht1: "\<Gamma>,(x:\<tau>') \<turnstile> ?e1x : \<tau>1 \<rightarrow> \<tau>"
      and Ht2: "\<Gamma>,(x:\<tau>') \<turnstile> ?e2x : \<tau>1" by auto
    have "\<Gamma> \<turnstile> vsubst_tm e' i e1 : \<tau>1 \<rightarrow> \<tau>" using "1.hyps" "1.prems" Hsize1 Hfree1 Ht1 by auto
    moreover
    have "\<Gamma> \<turnstile> vsubst_tm e' i e2 : \<tau>1" using "1.hyps" "1.prems" Hsize2 Hfree2 Ht2 by auto
    ultimately
    show ?thesis using TmApp by auto
  next
    case (TmNew e_b)
    let ?e_bx = "vsubst_tm (TmFreeVar x) i e_b"
    have "\<Gamma>,(x:\<tau>') \<turnstile> TmNew ?e_bx : \<tau>" using prems by auto
    then obtain \<tau>_b where Htb: "\<Gamma>,(x:\<tau>') \<turnstile> ?e_bx : \<tau>_b" and Heq: "\<tau> = TyRef \<tau>_b" by auto
    have "\<Gamma> \<turnstile> vsubst_tm e' i e_b : \<tau>_b" using "1.hyps" TmNew Htb "1.prems" by auto
    thus ?thesis using TmNew Heq by auto
  next
    case (TmAssign e1 e2)
    have Hsize1: "size e1 < n_y" and Hsize2: "size e2 < n_y" using "1.prems" TmAssign by auto
    have Hfree1: "\<not> x free in e1" and Hfree2: "\<not> x free in e2" using "1.prems" TmAssign by auto
    let ?e1x = "vsubst_tm (TmFreeVar x) i e1"
    let ?e2x = "vsubst_tm (TmFreeVar x) i e2"
    have "\<Gamma>,(x:\<tau>') \<turnstile> ?e1x := ?e2x : \<tau>" using prems by auto
    then obtain \<tau>_b where Ht1: "\<Gamma>,(x:\<tau>') \<turnstile> ?e1x : TyRef \<tau>_b"
      and Ht2: "\<Gamma>,(x:\<tau>') \<turnstile> ?e2x : \<tau>_b"
      and Heq: "\<tau> = TyUnit" by auto
    have "\<Gamma> \<turnstile> vsubst_tm e' i e1 : TyRef \<tau>_b" using "1.hyps" "1.prems" Hsize1 Hfree1 Ht1 by auto
    moreover
    have "\<Gamma> \<turnstile> vsubst_tm e' i e2 : \<tau>_b" using "1.hyps" "1.prems" Hsize2 Hfree2 Ht2 by auto
    ultimately
    show ?thesis using TmAssign Heq by auto
  next
    case (TmDeref e_b)
    let ?e_bx = "vsubst_tm (TmFreeVar x) i e_b"
    have Ht: "\<Gamma>,(x:\<tau>') \<turnstile> TmDeref ?e_bx : \<tau>" using prems by auto
    hence "\<Gamma>,(x:\<tau>') \<turnstile> ?e_bx : TyRef \<tau>" by auto
    hence "\<Gamma> \<turnstile> vsubst_tm e' i e_b : TyRef \<tau>" using "1.hyps" TmDeref "1.prems" by auto
    thus ?thesis using TmDeref by auto
  next
    case (TmLoc l)
    then obtain \<tau>_b where Heq: "\<tau> = TyRef \<tau>_b" using "1.prems" by auto
    hence "\<Gamma> \<turnstile> TmLoc l : TyRef \<tau>_b" using Hgok TmLoc "1.prems" by auto
    thus ?thesis using TmLoc Heq by auto
  qed
qed

constdefs
  storeTyping :: "store \<Rightarrow> env \<Rightarrow> bool"   ("\<Turnstile> _ : _" [50,50] 50)
  "storeTyping S \<Gamma> \<equiv> S \<Turnstile> Store \<and> (\<forall>x \<tau>. (x\<rhd>\<tau>) \<in> \<Gamma> \<longrightarrow> (\<exists>v. (x\<mapsto>v) \<in> S \<and> \<Gamma> \<turnstile> v : \<tau>))"

lemma store_ty_store:
  assumes "\<Turnstile> S : \<Gamma>"
  shows "S \<Turnstile> Store"
using prems by (auto simp add: storeTyping_def)

lemma store_ty_finite:
  assumes "\<Turnstile> S : \<Gamma>"
  shows "finite S"
using prems by (auto simp add: storeTyping_def wf_store_def)

lemma st_ty_update:
  assumes "\<Turnstile> S : \<Gamma>"
  and "(z\<rhd>\<tau>) \<in> \<Gamma>"
  and "\<Gamma> \<turnstile> v : \<tau>"
  and "value v"
  shows "\<Turnstile> update S z v : \<Gamma>"
proof (auto simp add: storeTyping_def)
  show "update S z v \<Turnstile> Store" using prems store_ty_store store_update_ok by auto
next
  fix xa \<tau>a
  assume "(xa\<rhd>\<tau>a) \<in> \<Gamma>"
  show "\<exists>va. (xa\<mapsto>va) \<in> update S z v \<and> \<Gamma> \<turnstile> va : \<tau>a" proof (cases "z = xa")
    case True
    have "(xa\<mapsto>v) \<in> update S z v" and "\<tau> = \<tau>a" using prems wf_env_def[of \<Gamma>] update_def by auto
    thus ?thesis using prems by auto
  next
    case False show ?thesis using prems by (auto simp add: update_def storeTyping_def)
  qed
qed

lemma store_dom_in:
  assumes "\<Turnstile> S : \<Gamma>"
  and "\<Gamma> defines z"
  shows "S defines z"
proof -
  obtain b where "(z,b) \<in> \<Gamma>" using prems by (auto simp add: dom_def)
  then obtain \<tau> where "(z\<rhd>\<tau>) \<in> \<Gamma>" by (cases b, auto)
  obtain v where "(z\<mapsto>v \<in> S)" using prems storeTyping_def[of S \<Gamma>] by fast
  thus ?thesis by (auto simp add: dom_def)
qed

lemma store_dom_not_in:
  assumes "\<Turnstile> S : \<Gamma>"
  and "\<not> S defines z"
  shows "\<not> \<Gamma> defines z"
using prems by (auto simp add: store_dom_in)

lemma store_typing_elim:
  assumes "\<Turnstile> S : \<Gamma>"
  and "(z\<mapsto>v) \<in> S"
  and "(z\<rhd>\<tau>) \<in> \<Gamma>"
  shows "\<Gamma> \<turnstile> v : \<tau>"
using prems storeTyping_def[of S \<Gamma>] wf_store_def[of S] by fast

lemma store_typing_cons:
  assumes "\<Turnstile> S : \<Gamma>"
  and "\<Gamma> \<turnstile> e : \<tau>"
  and "value e"
  and "\<not> S defines z"
  shows "\<Turnstile> S,(z\<mapsto>e) : \<Gamma>,(z:\<tau>)"
proof (unfold storeTyping_def, rule)
  show "S,(z\<mapsto>e) \<Turnstile> Store" using prems by (auto simp add: storeTyping_def wf_store_cons)
next
  have Hgok: "\<Gamma> \<Turnstile> OK" using prems ty_wf_env by auto
  have Hgnz: "\<not> \<Gamma> defines z" using prems store_dom_not_in by auto
  have Hgokz: "\<Gamma>,(z:\<tau>) \<Turnstile> OK" using Hgok Hgnz by (auto simp add: wf_env_cons)
  { fix xa \<tau>a
    assume H1: "(xa\<rhd>\<tau>a) \<in> \<Gamma>,(z:\<tau>)"
    have "\<exists>v. (xa\<mapsto>v) \<in> S,(z\<mapsto>e) \<and> \<Gamma>,(z:\<tau>) \<turnstile> v : \<tau>a" proof (cases "z=xa")
      case True
      hence "\<tau> = \<tau>a" using Hgokz H1 wf_env_def[of "\<Gamma>,(z:\<tau>)"] by auto
      hence "\<Gamma> \<turnstile> e : \<tau>a" using prems by auto
      hence "\<Gamma>,(z:\<tau>) \<turnstile> e : \<tau>a" using prems weakening[of _ _ _ "{z\<rhd>\<tau>}"] Hgokz by auto
      thus ?thesis using True by auto
    next
      case False
      hence "(xa\<rhd>\<tau>a) \<in> \<Gamma>" using H1 by auto
      then obtain v where "(xa\<mapsto>v) \<in> S \<and> \<Gamma> \<turnstile> v : \<tau>a" using storeTyping_def[of S \<Gamma>] prems by auto
      hence "(xa\<mapsto>v) \<in> S,(z\<mapsto>e) \<and> \<Gamma>,(z:\<tau>) \<turnstile> v : \<tau>a" using weakening[of _ _ _ "{z\<rhd>\<tau>}"] Hgokz by auto
      thus ?thesis by auto
    qed
  }
  thus "\<forall>x \<tau>'. x\<rhd>\<tau>' \<in> \<Gamma>,(z:\<tau>) \<longrightarrow> (\<exists>v. (x\<mapsto>v) \<in> S,(z\<mapsto>e) \<and> \<Gamma>,(z:\<tau>) \<turnstile> v : \<tau>')" by auto
qed

theorem preservation:
  assumes "\<Gamma> \<turnstile> e : \<tau>"
  and "S;e \<hookrightarrow> S';e'"
  and "\<Turnstile> S : \<Gamma>"
  shows "\<exists>\<Gamma>'. \<Turnstile> S' : \<Gamma> \<union> \<Gamma>' \<and> \<Gamma> \<union> \<Gamma>' \<turnstile> e' : \<tau>"
using prems proof (induct arbitrary: e')
  case (t_unit \<Gamma> e')
  have "S;TmUnit \<hookrightarrow> S';e'" .
  thus ?case proof cases
    case (e_val Sa v z)
    hence "\<Gamma>,(z:TyUnit) \<Turnstile> OK" using t_unit by (auto simp add: wf_env_cons store_dom_not_in)
    hence "\<Gamma>,(z:TyUnit) \<turnstile> TmFreeVar z : TyUnit" using t_unit by auto
    moreover
    have "\<Turnstile> S,(z\<mapsto>TmUnit) : \<Gamma>,(z:TyUnit)"
      using t_unit e_val store_typing_cons[of _ _ TmUnit TyUnit] by auto
    ultimately
    show ?thesis using e_val by (auto simp add: exI[of _ "{(z\<rhd>TyUnit)}"])
  qed auto
next
  case (t_var \<Gamma> \<tau> x e')
  have "S;TmFreeVar x \<hookrightarrow> S';e'" .
  thus ?case by (cases, auto)
next
  case (t_abs L \<Gamma> \<tau>1 \<tau>2 eb e')
  hence Ht: "\<Gamma> \<turnstile> \<lambda>[\<tau>1].eb : \<tau>1\<rightarrow>\<tau>2" by auto
  have "S;\<lambda>[\<tau>1].eb \<hookrightarrow> S';e'" .
  thus ?case proof cases
    case (e_val Sa v z)
    hence "\<Gamma>,(z:\<tau>1\<rightarrow>\<tau>2) \<Turnstile> OK" using t_abs by (auto simp add: wf_env_cons store_dom_not_in)
    hence "\<Gamma>,(z:\<tau>1\<rightarrow>\<tau>2) \<turnstile> TmFreeVar z : \<tau>1\<rightarrow>\<tau>2" using t_abs by auto
    moreover
    have "\<Turnstile> S,(z\<mapsto>\<lambda>[\<tau>1].eb) : \<Gamma>,(z:\<tau>1\<rightarrow>\<tau>2)"
      using e_val t_abs Ht store_typing_cons[of S \<Gamma> "\<lambda>[\<tau>1].eb" "\<tau>1\<rightarrow>\<tau>2" z] by auto
    ultimately
    show ?thesis using e_val by (auto simp add: exI[of _ "{(z\<rhd>\<tau>1\<rightarrow>\<tau>2)}"])
  qed auto
next
  case (t_app \<Gamma> \<tau>1 \<tau>2 e1 e2 e')
  have H1: "\<Gamma> \<turnstile> e1 : \<tau>1 \<rightarrow> \<tau>2" .
  have H2: "\<Gamma> \<turnstile> e2 : \<tau>1" .
  have "S;e1 \<cdot> e2 \<hookrightarrow> S';e'" .
  thus ?case using H1 H2 proof cases
    (* propagate 1 *)
    case (e_app_p1 Sa Sa' e1a e1')
    have "\<exists>\<Gamma>'. \<Turnstile> S' : \<Gamma> \<union> \<Gamma>' \<and> \<Gamma> \<union> \<Gamma>' \<turnstile> e1' : \<tau>1 \<rightarrow> \<tau>2" using t_app e_app_p1 by auto
    then obtain \<Gamma>' where "\<Turnstile> S' : \<Gamma> \<union> \<Gamma>' \<and> \<Gamma> \<union> \<Gamma>' \<turnstile> e1' : \<tau>1 \<rightarrow> \<tau>2" by auto
    moreover
    hence "\<Gamma> \<union> \<Gamma>' \<Turnstile> OK" by (auto simp add: ty_wf_env)
    hence "\<Gamma> \<union> \<Gamma>' \<turnstile> e2 : \<tau>1" using t_app by (auto simp add: weakening)
    ultimately
    show ?thesis using e_app_p1 by auto
  next
    (* propagate 2 *)
    case (e_app_p2 Sa Sa' e2a e2' z)
    have "\<exists>\<Gamma>'. \<Turnstile> S' : \<Gamma> \<union> \<Gamma>' \<and> \<Gamma> \<union> \<Gamma>' \<turnstile> e2' : \<tau>1" using t_app e_app_p2 by auto
    then obtain \<Gamma>' where "\<Turnstile> S' : \<Gamma> \<union> \<Gamma>' \<and> \<Gamma> \<union> \<Gamma>' \<turnstile> e2' : \<tau>1" by auto
    moreover
    hence "\<Gamma> \<union> \<Gamma>' \<Turnstile> OK" by (auto simp add: ty_wf_env)
    hence "\<Gamma> \<union> \<Gamma>' \<turnstile> TmFreeVar z : \<tau>1 \<rightarrow> \<tau>2" using t_app e_app_p2 by (auto simp add: weakening)
    ultimately
    show ?thesis using e_app_p2 by auto
  next
    (* beta *)
    case (e_beta Sa \<tau>1_y e1_y y w)
    hence Ht1: "\<Gamma> \<turnstile> \<lambda>[\<tau>1_y]. e1_y : \<tau>1 \<rightarrow> \<tau>2"
      using store_typing_elim[of Sa \<Gamma> _ "\<lambda>[\<tau>1_y].e1_y" "\<tau>1\<rightarrow>\<tau>2"] t_app e_beta by auto
    then obtain L where Hgok: "\<Gamma> \<Turnstile> OK" and Hfin: "finite L"
      and Hs: "\<forall>x. \<not> x free in e1_y \<and> x \<notin> L \<and> \<not> \<Gamma> defines x \<longrightarrow>
                 \<Gamma>,(x:\<tau>1_y) \<turnstile> freshen_tm x e1_y : \<tau>2"
      and Heq: "\<tau>1 = \<tau>1_y" by (rule typing_abs_elim, auto)
    have Ht2: "\<Gamma> \<turnstile> TmFreeVar y : \<tau>1_y" using e_beta t_app Heq by auto
    obtain z where Hz: "\<not> z free in e1_y \<and> z \<notin> L \<and> \<not> \<Gamma> defines z"
      using Hfin fresh_exists[of "set (free_tm e1_y) \<union> L \<union> dom \<Gamma>"]
        Hgok wf_env_def[of \<Gamma>] dom_finite[of \<Gamma>] by auto
    show ?thesis proof (rule exI[of _ "{}"], auto)
      have "\<Gamma>,(z:\<tau>1_y) \<turnstile> freshen_tm z e1_y : \<tau>2" using Hs Hz by auto
      hence "\<Gamma> \<turnstile> vsubst_tm (TmFreeVar y) 0 e1_y : \<tau>2"
        using Hz Hz Ht2 by (auto simp add: substitution)
      thus "\<Gamma> \<turnstile> e' : \<tau>2" using e_beta by auto
    next
      show "\<Turnstile> S' : \<Gamma>" using t_app e_beta by auto
    qed
  qed auto
next
  case (t_new \<Gamma> \<tau> e ef')
  have Ht: "\<Gamma> \<turnstile> e : \<tau>" .
  hence Hgok: "\<Gamma> \<Turnstile> OK" by (auto simp add: ty_wf_env)
  have "S; TmNew e \<hookrightarrow> S'; ef'" .
  thus ?case using Ht proof cases
    (* propagate *)
    case (e_new_p Sa Sa' ea ea')
    hence H: "\<exists>\<Gamma>'. \<Turnstile> S' : \<Gamma> \<union> \<Gamma>' \<and> \<Gamma> \<union> \<Gamma>' \<turnstile> ea': \<tau>" using prems by auto
    then obtain \<Gamma>' where Hst: "\<Turnstile> S': \<Gamma> \<union> \<Gamma>'" and Ht: "\<Gamma> \<union> \<Gamma>' \<turnstile> ea': \<tau>" by auto
    hence "\<Gamma> \<union> \<Gamma>' \<turnstile> TmNew ea': TyRef \<tau>" by auto
    thus ?thesis using t_new e_new_p Ht by auto
  next
    (* calculate *)
    case (e_new_c Sa v x z)
    hence "\<not> \<Gamma> defines z" using t_new by (auto simp add: store_dom_not_in)
    hence "\<Gamma>,(z:\<tau>) \<Turnstile> OK" using Hgok by (auto simp add: wf_env_cons)
    hence Htz: "\<Gamma>,(z:\<tau>) \<turnstile> TmLoc z : TyRef \<tau>" by auto
    have "S \<Turnstile> Store" using t_new by (auto simp add: store_ty_store)
    moreover
    hence "value v" using e_new_c by (auto simp add: store_ok_is_value)
    moreover
    have "\<Gamma> \<turnstile> v : \<tau>" using t_new e_new_c by (auto simp add: store_typing_elim)
    ultimately
    have "\<Turnstile> S,(z\<mapsto>v):\<Gamma>,(z:\<tau>)" using t_new e_new_c by (auto simp add: store_typing_cons)
    thus ?thesis using Htz e_new_c by (auto simp add: exI[of _ "{(z\<rhd>\<tau>)}"])
  qed auto
next
  case (t_assign \<Gamma> \<tau> e1 e2 ef')
  have H1: "\<Gamma> \<turnstile> e1 : TyRef \<tau>" .
  have H2: "\<Gamma> \<turnstile> e2 : \<tau>" .
  hence Hgok: "\<Gamma> \<Turnstile> OK" by (auto simp add: ty_wf_env)
  have "S; e1 := e2 \<hookrightarrow> S'; ef'" .
  thus ?case using H1 H2 proof cases
    (* propagate 1 *)
    case (e_assign_p1 Sa Sa' e1a e1')
    have "\<exists>\<Gamma>'. \<Turnstile> S' : \<Gamma> \<union> \<Gamma>' \<and> \<Gamma> \<union> \<Gamma>' \<turnstile> e1' : TyRef \<tau>" using t_assign e_assign_p1 by auto
    then obtain \<Gamma>' where "\<Turnstile> S' : \<Gamma> \<union> \<Gamma>' \<and> \<Gamma> \<union> \<Gamma>' \<turnstile> e1' : TyRef \<tau>" by auto
    moreover
    hence "\<Gamma> \<union> \<Gamma>' \<Turnstile> OK" by (auto simp add: ty_wf_env)
    hence "\<Gamma> \<union> \<Gamma>' \<turnstile> e2 : \<tau>" using t_assign by (auto simp add: weakening)
    ultimately
    show ?thesis using e_assign_p1 by auto
  next
    (* propagate 2 *)
    case (e_assign_p2 Sa Sa' e2a e2' z)
    have "\<exists>\<Gamma>'. \<Turnstile> S' : \<Gamma> \<union> \<Gamma>' \<and> \<Gamma> \<union> \<Gamma>' \<turnstile> e2' : \<tau>" using t_assign e_assign_p2 by auto
    then obtain \<Gamma>' where "\<Turnstile> S' : \<Gamma> \<union> \<Gamma>' \<and> \<Gamma> \<union> \<Gamma>' \<turnstile> e2' : \<tau>" by auto
    moreover
    hence "\<Gamma> \<union> \<Gamma>' \<Turnstile> OK" by (auto simp add: ty_wf_env)
    hence "\<Gamma> \<union> \<Gamma>' \<turnstile> TmFreeVar z : TyRef \<tau>" using t_assign e_assign_p2 by (auto simp add: weakening)
    ultimately
    show ?thesis using e_assign_p2 by auto
  next
    (* calculate *)
    case (e_assign_c Sa l v x y)
    hence Ht1: "\<Gamma> \<turnstile> TmLoc l : TyRef \<tau>"
      using store_typing_elim[of Sa \<Gamma> _ "TmLoc l" "TyRef \<tau>"] t_assign e_assign_c by auto
    hence Hloc: "(l\<rhd>\<tau>) \<in> \<Gamma>" by auto
    have Ht2: "\<Gamma> \<turnstile> TmFreeVar y : \<tau>" using e_assign_c t_assign by auto
    show ?thesis proof (rule exI[of _ "{}"], auto)
      have "S \<Turnstile> Store" using t_assign by (auto simp add: store_ty_store)
      moreover
      hence "value v" using e_assign_c by (auto simp add: store_ok_is_value)
      moreover
      have "\<Gamma> \<turnstile> v : \<tau>" using t_assign e_assign_c by (auto simp add: store_typing_elim)
      ultimately
      have "\<Turnstile> update S l v : \<Gamma>" using t_assign Hloc by (auto simp add: st_ty_update dom_def)
      thus "\<Turnstile> S' : \<Gamma>" using e_assign_c by auto
    next
      show "\<Gamma> \<turnstile> ef' : TyUnit" using Hgok e_assign_c by auto
    qed
  qed auto
next
  case (t_deref \<Gamma> \<tau> e ef')
  have Ht: "\<Gamma> \<turnstile> e : TyRef \<tau>" .
  hence Hgok: "\<Gamma> \<Turnstile> OK" by (auto simp add: ty_wf_env)
  have "S; TmDeref e \<hookrightarrow> S'; ef'" .
  thus ?case proof cases
    (* propagate *)
    case (e_deref_p Sa Sa' ea ea')
    hence H: "\<exists>\<Gamma>'. \<Turnstile> S' : \<Gamma> \<union> \<Gamma>' \<and> \<Gamma> \<union> \<Gamma>' \<turnstile> ea': TyRef \<tau>" using prems by auto
    then obtain \<Gamma>' where Hst: "\<Turnstile> S': \<Gamma> \<union> \<Gamma>'" and Ht': "\<Gamma> \<union> \<Gamma>' \<turnstile> ea': TyRef \<tau>" by auto
    hence "\<Gamma> \<union> \<Gamma>' \<turnstile> TmDeref ea': \<tau>" by auto
    thus ?thesis using t_deref e_deref_p Ht' by auto
  next
    (* calculate *)
    case (e_deref_c Sa l v x)
    hence "(x\<rhd>TyRef \<tau>) \<in> \<Gamma>" using Ht by auto
    hence "\<Gamma> \<turnstile> TmLoc l : TyRef \<tau>" using t_deref e_deref_c by (auto simp add: store_typing_elim)
    hence "\<Gamma> \<turnstile> v : \<tau>"  using t_deref e_deref_c store_typing_elim[of S \<Gamma> l v \<tau>] by auto
    thus ?thesis using t_deref e_deref_c by (auto simp add: exI[of _ "{}"])
  qed auto
next
  case (t_loc \<Gamma> \<tau> l e')
  have "S;TmLoc l \<hookrightarrow> S';e'" .
  thus ?case proof cases
    case (e_val Sa v z)
    hence "\<Gamma>,(z:TyRef \<tau>) \<Turnstile> OK" using t_loc by (auto simp add: wf_env_cons store_dom_not_in)
    hence "\<Gamma>,(z:TyRef \<tau>) \<turnstile> TmFreeVar z : TyRef \<tau>" using t_loc by auto
    moreover
    have "\<Turnstile> S,(z\<mapsto>TmLoc l) : \<Gamma>,(z:TyRef \<tau>)"
      using t_loc e_val store_typing_cons[of _ _ "TmLoc l" "TyRef \<tau>"] by auto
    ultimately
    show ?thesis using e_val by (auto simp add: exI[of _ "{(z\<rhd>TyRef \<tau>)}"])
  qed auto
qed

theorem progress:
  assumes "\<Gamma> \<turnstile> e : \<tau>"
  and "\<Turnstile> S : \<Gamma>"
  shows "not_stuck e S"
proof -
  have Hstok: "S \<Turnstile> Store" using prems by (auto simp add: store_ty_store)
  show ?thesis using prems proof induct
    case (t_unit \<Gamma>)
    obtain z where "\<not> S defines z"
      using prems fresh_exists[of "dom S"] by (auto simp add: dom_finite store_ty_finite)
    hence "S;TmUnit \<hookrightarrow> S,(z\<mapsto>TmUnit);TmFreeVar z" using Hstok by auto
    thus ?case by auto
  next
    case (t_abs L \<Gamma> \<tau>1 \<tau>2 e_b)
    obtain z where "\<not> S defines z"
      using prems fresh_exists[of "dom S"] by (auto simp add: dom_finite store_ty_finite)
    hence "S;\<lambda>[\<tau>1]. e_b \<hookrightarrow> S,(z\<mapsto>\<lambda>[\<tau>1]. e_b);TmFreeVar z" using Hstok by auto
    thus ?case by auto
  next
    case (t_app \<Gamma> \<tau>1 \<tau>2 e1 e2)
    hence "isVar e1 \<or> makes_progress e1 S" by auto
    thus ?case proof
      assume "isVar e1"
      then obtain x where Hx: "e1 = TmFreeVar x" by auto
      have "isVar e2 \<or> makes_progress e2 S" using t_app by auto
      thus ?case proof
        assume "isVar e2"
        then obtain y where Hy: "e2 = TmFreeVar y" by auto
        have "\<Gamma> defines x" using Hx t_app by (auto simp add: dom_def)
        hence "S defines x" using t_app by (auto simp add: store_dom_in)
        then obtain v where Hxv: "(x\<mapsto>v) \<in> S" and Hv: "value v"
          using Hstok store_ok_exists_value by auto
        hence "\<Gamma> \<turnstile> v : \<tau>1 \<rightarrow> \<tau>2" using t_app Hx by (auto simp add: store_typing_elim)
        then obtain e_b where "v = \<lambda>[\<tau>1]. e_b" using Hv by auto
        hence "S;x\<Down>\<lambda>[\<tau>1]. e_b" using Hxv by auto
        hence "S;TmFreeVar x \<cdot> TmFreeVar y \<hookrightarrow> S;freshen_tm y e_b" ..
        thus ?thesis using Hx Hy by force
      next
        assume "makes_progress e2 S"
        thus ?thesis using Hx by auto
      qed
    next
      assume "makes_progress e1 S"
      thus ?thesis by auto
    qed
  next
    case (t_new \<Gamma> \<tau> e)
    hence "isVar e \<or> makes_progress e S" by auto
    thus ?case proof
      assume "isVar e"
      then obtain x where Hx: "e = TmFreeVar x" by auto
      have "\<Gamma> defines x" using Hx t_new by (auto simp add: dom_def)
      hence "S defines x" using t_new by (auto simp add: store_dom_in)
      then obtain v where Hxv: "(x\<mapsto>v) \<in> S" and Hv: "value v"
        using Hstok store_ok_exists_value by auto
      obtain z where "\<not> S defines z"
        using prems fresh_exists[of "dom S"] by (auto simp add: dom_finite store_ty_finite)
      hence "S;TmNew (TmFreeVar x) \<hookrightarrow> S,(z\<mapsto>v);TmLoc z" using prems Hx by auto
      thus ?thesis using Hx by force
    next
      assume "makes_progress e S"
      thus ?thesis by auto
    qed
  next
    case (t_assign \<Gamma> \<tau> e1 e2)
    hence "isVar e1 \<or> makes_progress e1 S" by auto
    thus ?case proof
      assume "isVar e1"
      then obtain x where Hx: "e1 = TmFreeVar x" by auto
      have "isVar e2 \<or> makes_progress e2 S" using t_assign by auto
      thus ?case proof
        assume "isVar e2"
        then obtain y where Hy: "e2 = TmFreeVar y" by auto
        hence "\<Gamma> defines y" using t_assign by (auto simp add: dom_def)
        hence "S defines y" using t_assign by (auto simp add: store_dom_in)
        then obtain v' where "(y\<mapsto>v') \<in> S" and "value v'"
          using Hstok store_ok_exists_value by auto
        hence "S;y\<Down>v'" by auto
        moreover
        have "\<Gamma> defines x" using Hx t_assign by (auto simp add: dom_def)
        hence "S defines x" using t_assign by (auto simp add: store_dom_in)
        then obtain v where Hxv: "(x\<mapsto>v) \<in> S" and Hv: "value v"
          using Hstok store_ok_exists_value by auto
        hence "\<Gamma> \<turnstile> v : TyRef \<tau>" using t_assign Hx by (auto simp add: store_typing_elim)
        then obtain l where "v = TmLoc l" using Hv by auto
        hence "S;x\<Down>TmLoc l" using Hxv by auto
        ultimately
        have "S;TmFreeVar x := TmFreeVar y \<hookrightarrow> update S l v';TmUnit" by auto
        thus ?thesis using Hx Hy by force
      next
        assume "makes_progress e2 S"
        thus ?thesis using Hx by auto
      qed
    next
      assume "makes_progress e1 S"
      thus ?thesis by auto
    qed
  next
    case (t_deref \<Gamma> \<tau> e)
    hence "isVar e \<or> makes_progress e S" by auto
    thus ?case proof
      assume "isVar e"
      then obtain x where Hx: "e = TmFreeVar x" by auto
      have "\<Gamma> defines x" using Hx t_deref by (auto simp add: dom_def)
      hence "S defines x" using t_deref by (auto simp add: store_dom_in)
      then obtain v where Hxv: "(x\<mapsto>v) \<in> S" and Hv: "value v"
        using Hstok store_ok_exists_value by auto
      hence Htv: "\<Gamma> \<turnstile> v : TyRef \<tau>" using t_deref Hx by (auto simp add: store_typing_elim)
      then obtain l where Heqv: "v = TmLoc l" using Hv by auto
      hence "S;x\<Down>TmLoc l" using Hxv by auto
      moreover
      have "\<Gamma> defines l" using Htv Heqv by (auto simp add:dom_def)
      then obtain v' where "S;l\<Down>v'"
        using t_deref store_dom_in Hstok store_ok_exists_value[of S l] by auto
      ultimately
      have "S;TmDeref (TmFreeVar x) \<hookrightarrow> S;v'" .. 
      thus ?thesis using Hx by force
    next
      assume "makes_progress e S"
      thus ?thesis by auto
    qed
  next
    case (t_loc \<Gamma> \<tau> l)
    obtain z where "\<not> S defines z"
      using prems fresh_exists[of "dom S"] by (auto simp add: dom_finite store_ty_finite)
    hence "S;TmLoc l \<hookrightarrow> S,(z\<mapsto>TmLoc l);TmFreeVar z" using Hstok by auto
    thus ?case by auto
  qed auto
qed

theorem safety:
  assumes Ht: "\<Gamma> \<turnstile> e : \<tau>"
  and Hsem: "S;e \<hookrightarrow>* S';e'"
  and Hstok: "\<Turnstile> S : \<Gamma>"
  shows "not_stuck e' S'"
using Hsem Ht Hstok proof (induct arbitrary: \<Gamma>)
  case e_refl thus ?case using progress by auto
next
  case e_trans thus ?case using progress preservation by blast
qed

theorem program_safety:
  assumes "{} \<turnstile> e : \<tau>"
  and "{};e \<hookrightarrow>* S';e'"
  shows "not_stuck e' S'"
using prems by (rule safety, auto simp add: storeTyping_def wf_store_def)

end
