theory Syntax
imports Environ

begin

datatype ty =
    TyUnit
  | TyBase name
  | TyFreeVar name
  | TyVar nat
  | TyArrow ty ty    (infixr "\<rightarrow>" 900)
  | TyForall ty      ("\<forall>._" [900] 900)
  | TyRef ty

datatype tm =
    TmUnit
  | TmFreeVar name
  | TmVar nat
  | TmAbs ty tm       ("\<lambda>[_]._" [900,800] 800)
  | TmApp tm tm       (infixl "\<cdot>" 900)
  | TmProd tm         ("\<Lambda>[*]._" [900] 900)
  | TmTapp tm ty      ("_\<langle>_\<rangle>" [900,0] 900)
  | TmNew tm
  | TmAssign tm tm    (infixr ":=" 900)
  | TmDeref tm
  | TmLoc name

(* vsubst: freshening substitution *)
consts vsubst_tm:: "tm \<Rightarrow> nat \<Rightarrow> tm \<Rightarrow> tm"
primrec
  "vsubst_tm s i TmUnit = TmUnit"
  "vsubst_tm s i (TmFreeVar b) = TmFreeVar b"
  "vsubst_tm s i (TmVar n) = (if i = n then s else (TmVar n))"
  "vsubst_tm s i (\<lambda>[\<tau>].e) = \<lambda>[\<tau>].(vsubst_tm s (Suc i) e)"
  "vsubst_tm s i (e1\<cdot>e2) = (vsubst_tm s i e1)\<cdot>(vsubst_tm s i e2)"
  "vsubst_tm s i (\<Lambda>[*].e) = \<Lambda>[*]. (vsubst_tm s i e)"
  "vsubst_tm s i (e\<langle>\<tau>\<rangle>) = (vsubst_tm s i e)\<langle>\<tau>\<rangle>"
  "vsubst_tm s i (TmNew e) = TmNew (vsubst_tm s i e)"
  "vsubst_tm s i (TmAssign e1 e2) = TmAssign (vsubst_tm s i e1) (vsubst_tm s i e2)"
  "vsubst_tm s i (TmDeref e) = TmDeref (vsubst_tm s i e)"
  "vsubst_tm s i (TmLoc l) = TmLoc l"

consts vsubst_ty:: "ty \<Rightarrow> nat \<Rightarrow> ty \<Rightarrow> ty"
primrec
  "vsubst_ty s i TyUnit = TyUnit"
  "vsubst_ty s i (TyBase n) = (TyBase n)"
  "vsubst_ty s i (TyFreeVar b) = TyFreeVar b"
  "vsubst_ty s i (TyVar n) = (if n=i then s else (TyVar n))"
  "vsubst_ty s i (\<tau>1\<rightarrow>\<tau>2) = (vsubst_ty s i \<tau>1) \<rightarrow> (vsubst_ty s i \<tau>2)"
  "vsubst_ty s i (\<forall>.\<tau>) = \<forall>.(vsubst_ty s (Suc i) \<tau>)"
  "vsubst_ty s i (TyRef \<tau>) = TyRef (vsubst_ty s i \<tau>)"

consts convert_ty:: "nat \<Rightarrow> nat \<Rightarrow> ty \<Rightarrow> ty"
primrec
  "convert_ty j i TyUnit = TyUnit"
  "convert_ty j i (TyBase n) = (TyBase n)"
  "convert_ty j i (TyFreeVar b) = TyFreeVar b"
  "convert_ty j i (TyVar n) = (if n=i then (TyVar j) else (TyVar n))"
  "convert_ty j i (\<tau>1\<rightarrow>\<tau>2) = (convert_ty j i \<tau>1) \<rightarrow> (convert_ty j i \<tau>2)"
  "convert_ty j i (\<forall>.\<tau>) = \<forall>.(convert_ty (Suc j) (Suc i) \<tau>)"
  "convert_ty j i (TyRef \<tau>) = TyRef (convert_ty j i \<tau>)"

consts vsubst_tmty:: "ty \<Rightarrow> nat \<Rightarrow> tm \<Rightarrow> tm"
primrec
  "vsubst_tmty s i TmUnit = TmUnit"
  "vsubst_tmty s i (TmFreeVar b) = TmFreeVar b"
  "vsubst_tmty s i (TmVar n) = (TmVar n)"
  "vsubst_tmty s i (\<lambda>[\<tau>].e) = \<lambda>[vsubst_ty s i \<tau>].(vsubst_tmty s i e)"
  "vsubst_tmty s i (e1\<cdot>e2) = (vsubst_tmty s i e1)\<cdot>(vsubst_tmty s i e2)"
  "vsubst_tmty s i (\<Lambda>[*].e) = (\<Lambda>[*].(vsubst_tmty s (Suc i) e))"
  "vsubst_tmty s i (e\<langle>\<tau>\<rangle>) = (vsubst_tmty s i e)\<langle>vsubst_ty s i \<tau>\<rangle>"
  "vsubst_tmty s i (TmNew e) = TmNew (vsubst_tmty s i e)"
  "vsubst_tmty s i (TmAssign e1 e2) = TmAssign (vsubst_tmty s i e1) (vsubst_tmty s i e2)"
  "vsubst_tmty s i (TmDeref e) = TmDeref (vsubst_tmty s i e)"
  "vsubst_tmty s i (TmLoc l) = TmLoc l"

consts
  freshen_tm :: "name \<Rightarrow> tm \<Rightarrow> tm"
  freshen_ty :: "name \<Rightarrow> ty \<Rightarrow> ty"
  freshen_tmty :: "name \<Rightarrow> tm \<Rightarrow> tm"

defs freshen_tm[simp]: "freshen_tm x e \<equiv> vsubst_tm (TmFreeVar x) 0 e"
defs freshen_ty[simp]: "freshen_ty x \<tau> \<equiv> vsubst_ty (TyFreeVar x) 0 \<tau>"
defs freshen_tmty[simp]: "freshen_tmty x e \<equiv> vsubst_tmty (TyFreeVar x) 0 e"

consts free_ty :: "ty \<Rightarrow> name list"
primrec
  "free_ty (TyUnit) = []"
  "free_ty (TyBase n) = []"
  "free_ty (TyFreeVar a) = [a]"
  "free_ty (TyVar n) = []"
  "free_ty (\<tau>1\<rightarrow>\<tau>2) = free_ty \<tau>1 @ free_ty \<tau>2"
  "free_ty (\<forall>.\<tau>) = free_ty \<tau>"
  "free_ty (TyRef \<tau>) = free_ty \<tau>"

consts free_tm :: "tm \<Rightarrow> name list"
primrec
  "free_tm TmUnit           = []"
  "free_tm (TmFreeVar b)    = [b]"
  "free_tm (TmVar n)        = []"
  "free_tm (\<lambda>[\<tau>].e)         = free_tm e"
  "free_tm (e1\<cdot>e2)          = free_tm e1 @ free_tm e2"
  "free_tm (\<Lambda>[*].e)         = free_tm e"
  "free_tm (e\<langle>\<tau>\<rangle>)            = free_tm e"
  "free_tm (TmNew e)        = free_tm e"
  "free_tm (TmAssign e1 e2) = free_tm e1 @ free_tm e2"
  "free_tm (TmDeref e)      = free_tm e"
  "free_tm (TmLoc l)        = [l]"

consts free_tmty :: "tm \<Rightarrow> name list"
primrec
  "free_tmty TmUnit           = []"
  "free_tmty (TmFreeVar b)    = []"
  "free_tmty (TmVar n)        = []"
  "free_tmty (\<lambda>[\<tau>].e)         = free_ty \<tau> @ free_tmty e"
  "free_tmty (e1\<cdot>e2)          = free_tmty e1 @ free_tmty e2"
  "free_tmty (\<Lambda>[*].e)         = free_tmty e"
  "free_tmty (e\<langle>\<tau>\<rangle>)            = free_tmty e @ free_ty \<tau>"
  "free_tmty (TmNew e)        = free_tmty e"
  "free_tmty (TmAssign e1 e2) = free_tmty e1 @ free_tmty e2"
  "free_tmty (TmDeref e)      = free_tmty e"
  "free_tmty (TmLoc l)        = []"

syntax
  is_free_in_ty :: "name \<Rightarrow> ty \<Rightarrow> bool"  ("_ free in type _" [900,900] 91)
  is_free_in_tm :: "name \<Rightarrow> tm \<Rightarrow> bool"  ("_ free in _" [900,900] 91)
  is_free_in_tmty :: "name \<Rightarrow> tm \<Rightarrow> bool"  ("_ free in term _" [900,900] 91)
translations
  "x free in type \<tau>" \<rightleftharpoons> "x \<in> set (free_ty \<tau>)"
  "x free in e" \<rightleftharpoons> "x \<in> set (free_tm e)"
  "x free in term e" \<rightleftharpoons> "x \<in> set (free_tmty e)"

consts shift_many_db_list :: "nat \<Rightarrow> nat list \<Rightarrow> nat list"
primrec
  "shift_many_db_list k [] = []"
  "shift_many_db_list k (n#l) = (if n < k then shift_many_db_list k l
                                          else (n - k)#(shift_many_db_list k l))"

consts shift_db_list :: "nat list \<Rightarrow> nat list"
defs shift_db_list[simp]: "shift_db_list l \<equiv> shift_many_db_list 1 l"

lemma shift_many_sanity:
  shows "set (shift_many_db_list k l) = { x-k | x. x \<ge> k \<and> x \<in> set l }"
by (induct l, auto)

consts free_db_ty :: "ty \<Rightarrow> nat list"
primrec
  "free_db_ty (TyUnit) = []"
  "free_db_ty (TyBase n) = []"
  "free_db_ty (TyFreeVar a) = []"
  "free_db_ty (TyVar n) = [n]"
  "free_db_ty (\<tau>1\<rightarrow>\<tau>2) = free_db_ty \<tau>1 @ free_db_ty \<tau>2"
  "free_db_ty (\<forall>.\<tau>) = shift_db_list (free_db_ty \<tau>)"
  "free_db_ty (TyRef \<tau>) = free_db_ty \<tau>"

consts free_db_tm :: "tm \<Rightarrow> nat list"
primrec
  "free_db_tm TmUnit           = []"
  "free_db_tm (TmFreeVar b)    = []"
  "free_db_tm (TmVar n)        = [n]"
  "free_db_tm (\<lambda>[\<tau>]. e)        = shift_db_list (free_db_tm e)"
  "free_db_tm (e1\<cdot>e2)          = free_db_tm e1 @ free_db_tm e2"
  "free_db_tm (\<Lambda>[*].e)         = free_db_tm e"
  "free_db_tm (e\<langle>\<tau>\<rangle>)            = free_db_tm e"
  "free_db_tm (TmNew e)        = free_db_tm e"
  "free_db_tm (TmAssign e1 e2) = free_db_tm e1 @ free_db_tm e2"
  "free_db_tm (TmDeref e)      = free_db_tm e"
  "free_db_tm (TmLoc l)        = []"

consts free_db_tmty :: "tm \<Rightarrow> nat list"
primrec
  "free_db_tmty TmUnit           = []"
  "free_db_tmty (TmFreeVar b)    = []"
  "free_db_tmty (TmVar n)        = []"
  "free_db_tmty (\<lambda>[\<tau>].e)         = free_db_ty \<tau> @ free_db_tmty e"
  "free_db_tmty (e1\<cdot>e2)          = free_db_tmty e1 @ free_db_tmty e2"
  "free_db_tmty (\<Lambda>[*].e)         = shift_db_list (free_db_tmty e)"
  "free_db_tmty (e\<langle>\<tau>\<rangle>)            = free_db_tmty e @ free_db_ty \<tau>"
  "free_db_tmty (TmNew e)        = free_db_tmty e"
  "free_db_tmty (TmAssign e1 e2) = free_db_tmty e1 @ free_db_tmty e2"
  "free_db_tmty (TmDeref e)      = free_db_tmty e"
  "free_db_tmty (TmLoc l)        = []"

lemma not_free_subst_ty:
  assumes "\<not> a free in type \<tau>"
  and "\<not> a free in type \<tau>'"
  shows "\<not> a free in type (vsubst_ty \<tau>' k \<tau>)"
using prems by (induct \<tau> arbitrary: k, auto)

lemma free_subst_ty:
  assumes "a free in type \<tau>"
  shows "a free in type (vsubst_ty \<tau>' k \<tau>)"
using prems by (induct \<tau> arbitrary: k, auto)

lemma inv_not_free_subst_ty:
  assumes "\<not> a free in type (vsubst_ty \<tau>' k \<tau>)"
  shows "\<not> a free in type \<tau>"
using prems free_subst_ty by auto

lemma not_free_subst:
  assumes "\<not> x free in e"
  and "\<not> x free in e'"
  shows "\<not> x free in (vsubst_tm e' k e)"
using prems by (induct e arbitrary: k, auto)

lemma not_free_subst_tm:
  assumes "\<not> a free in term e"
  and "\<not> a free in term e'"
  shows "\<not> a free in term (vsubst_tm e' k e)"
using prems by (induct e arbitrary: k, auto)

lemma not_free_subst_sty:
  assumes "\<not> x free in e"
  shows "\<not> x free in (vsubst_tmty \<tau> k e)"
using prems by (induct e arbitrary: k, auto)

lemma not_free_subst_tmty:
  assumes "\<not> a free in term e"
  and "\<not> a free in type \<tau>"
  shows "\<not> a free in term (vsubst_tmty \<tau> k e)"
using prems by (induct e arbitrary: k, auto simp add: not_free_subst_ty)

lemma not_free_convert_ty:
  assumes "\<not> a free in type \<tau>"
  shows "\<not> a free in type convert_ty j i \<tau>"
using prems by (induct \<tau> arbitrary: i j, auto)

lemma vsubst_ty_var_same_size[simp]:
  shows "size (vsubst_ty (TyFreeVar x) k \<tau>) = size \<tau>"
by (induct \<tau> arbitrary: k, auto)

lemma vsubst_var_same_size[simp]:
  shows "size (vsubst_tm (TmFreeVar x) k e) = size e"
by (induct e arbitrary: k, auto)

lemma vsubst_tmty_var_same_size[simp]:
  shows "size (vsubst_tmty (TyFreeVar x) k e) = size e"
by (induct e arbitrary: k, auto)

lemma subst_subst_ty_swap:
  assumes "i \<noteq> j"
  and "\<forall>k. vsubst_ty (TyFreeVar a) k \<tau>_s = \<tau>_s"
  shows "vsubst_ty (TyFreeVar a) i (vsubst_ty \<tau>_s j \<tau>) =
         vsubst_ty \<tau>_s j (vsubst_ty (TyFreeVar a) i \<tau>)"
using prems by (induct \<tau> arbitrary: i j, auto)

lemma subst_subst_swap:
  assumes "i \<noteq> j"
  and "\<forall>k. vsubst_tm (TmFreeVar x) k e_s = e_s"
  shows "vsubst_tm (TmFreeVar x) i (vsubst_tm e_s j e) =
         vsubst_tm e_s j (vsubst_tm (TmFreeVar x) i e)"
using prems by (induct e arbitrary: i j, auto)

lemma subst_subst_tmty_swap:
  assumes "\<forall>k. vsubst_tmty (TyFreeVar a) k e = e"
  shows "vsubst_tm e i (vsubst_tmty (TyFreeVar a) j eb) =
         vsubst_tmty (TyFreeVar a) j (vsubst_tm e i eb)"
using prems by (induct eb arbitrary: i j, auto)

lemma subst_subst_tytm_swap:
  shows "vsubst_tm (TmFreeVar x) i (vsubst_tmty \<tau> j eb) =
         vsubst_tmty \<tau> j (vsubst_tm (TmFreeVar x) i eb)"
by (induct eb arbitrary: i j, auto)

lemma subst_subst_tyty_swap:
  assumes "i \<noteq> j"
  and "\<forall>k. vsubst_ty (TyFreeVar a) k \<tau> = \<tau>"
  shows "vsubst_tmty \<tau> i (vsubst_tmty (TyFreeVar a) j eb) =
         vsubst_tmty (TyFreeVar a) j (vsubst_tmty \<tau> i eb)"
using prems by (induct eb arbitrary: i j, auto simp add: subst_subst_ty_swap)

lemma subst_convert_subst_swap:
  assumes "i \<noteq> k"
  and "j \<noteq> k"
  and "\<forall>k. vsubst_ty (TyFreeVar a) k \<tau>' = \<tau>'"
  shows "vsubst_ty \<tau>' i (convert_ty i j (vsubst_ty (TyFreeVar a) k \<tau>)) =
         vsubst_ty (TyFreeVar a) k (vsubst_ty \<tau>' i (convert_ty i j \<tau>))"
using prems by (induct \<tau> arbitrary: i j k, auto)

lemma vsubst_ty_free_db_ty:
  assumes "free_db_ty \<tau>_s = []"
  shows "set (free_db_ty (vsubst_ty \<tau>_s k \<tau>)) = set (free_db_ty \<tau>) - {k}"
using prems by (induct \<tau> arbitrary: k, auto simp add: shift_many_sanity)

lemma vsubst_tm_free_db_tm:
  assumes "free_db_tm e_s = []"
  shows "set (free_db_tm (vsubst_tm e_s k e)) = set (free_db_tm e) - {k}"
using prems by (induct e arbitrary: k, auto simp add: shift_many_sanity)

lemma vsubst_tmty_free_db:
  shows "free_db_tm e = free_db_tm (vsubst_tmty \<tau> i e)"
by (induct e arbitrary: i, auto)

lemma vsubst_tmty_free_db_tmty:
  assumes "free_db_ty \<tau>_s = []"
  shows "set (free_db_tmty (vsubst_tmty \<tau>_s k e)) = set (free_db_tmty e) - {k}"
using prems by (induct e arbitrary: k, auto simp add: shift_many_sanity vsubst_ty_free_db_ty)

lemma vsubst_tmty_free_db_tm:
  assumes "free_db_tmty e_s = []"
  shows "set (free_db_tmty (vsubst_tm e_s k e)) = set (free_db_tmty e)"
using prems by (induct e arbitrary: k, auto simp add: shift_many_sanity)

lemma not_in_free_db_same_subst_ty:
  assumes "k \<notin> set (free_db_ty \<tau>)"
  shows "vsubst_ty \<tau>' k \<tau> = \<tau>"
using prems proof (induct \<tau> arbitrary: k)
  case (TyForall \<tau>b k')
  thus ?case using shift_many_sanity[of "1" "free_db_ty \<tau>b"] by auto
qed auto

lemma not_in_free_db_same_subst:
  assumes "k \<notin> set (free_db_tm e)"
  shows "vsubst_tm e' k e = e"
using prems proof (induct e arbitrary: k)
  case (TmAbs \<tau> eb k')
  thus ?case using shift_many_sanity[of "1" "free_db_tm eb"] by auto
qed auto

lemma not_in_free_db_same_subst_tmty:
  assumes "k \<notin> set (free_db_tmty e)"
  shows "vsubst_tmty \<tau>' k e = e"
using prems proof (induct e arbitrary: k)
  case (TmProd eb k')
  thus ?case using shift_many_sanity[of "1" "free_db_tmty eb"] 
    by (auto simp add: not_in_free_db_same_subst_ty)
qed (auto simp add: not_in_free_db_same_subst_ty)

lemma no_free_db_vars_same_subst_ty:
  assumes "free_db_ty \<tau> = []"
  shows "vsubst_ty \<tau>' k \<tau> = \<tau>"
using prems by (auto simp add: not_in_free_db_same_subst_ty)

lemma no_free_db_vars_same_subst:
  assumes "free_db_tm e = []"
  shows "vsubst_tm e' k e = e"
using prems by (auto simp add: not_in_free_db_same_subst)

lemma no_free_db_vars_same_subst_tmty:
  assumes "free_db_tmty e = []"
  shows "vsubst_tmty \<tau>' k e = e"
using prems by (auto simp add: not_in_free_db_same_subst_tmty)

consts Val :: "tm set"
syntax
  "_Val" :: "tm \<Rightarrow> bool"   ("value")
translations
  "value e" \<rightleftharpoons> "e \<in> Val"

inductive Val
  intros
  v_unit[intro!]  : "value TmUnit"
  v_abs[intro!]   : "value (\<lambda>[\<tau>].e)"
  v_poly[intro!]  : "\<lbrakk> value v \<rbrakk> \<Longrightarrow> value (\<Lambda>[*].v)"
  v_loc[intro!]   : "value (TmLoc l)"

inductive_cases value_elim[elim]: "value e"

lemma value_subst_value:
  assumes "value (vsubst_tm (TmFreeVar x) i e)"
  shows "value e"
using prems proof (induct e, auto)
  fix n
  assume "value (if i = n then TmFreeVar x else TmVar n)"
  show "value (TmVar n)" proof (cases "i = n")
    case True
    have "value (TmFreeVar x)" using prems by auto
    thus ?thesis by auto
  next
    case False
    have "value (TmVar n)" using prems by auto
    thus ?thesis by auto
  qed
qed

lemma value_subst_value_rev:
  assumes "value e"
  shows "value (vsubst_tm e' i e)"
using prems by (induct, auto)

lemma subst_subst_equal_tyty:
  assumes "vsubst_ty (TyFreeVar x) i \<tau>1 = vsubst_ty (TyFreeVar x) i \<tau>2"
  and "\<not> x free in type \<tau>1"
  and "\<not> x free in type \<tau>2"
  shows "\<tau>1 = \<tau>2"
using prems proof (induct \<tau>1 arbitrary: i \<tau>2)
  case (TyUnit i \<tau>2)
  hence "TyUnit = vsubst_ty (TyFreeVar x) i \<tau>2" by auto
  thus ?case proof (cases \<tau>2, auto)
    fix n
    assume "TyUnit = (if n = i then TyFreeVar x else TyVar n)"
    thus False by (cases "n=i", auto)
  qed
next
  case (TyBase nam i \<tau>2)
  hence "TyBase nam = vsubst_ty (TyFreeVar x) i \<tau>2" by auto
  thus ?case by (cases \<tau>2, auto)
next
  case (TyFreeVar z i \<tau>2)
  hence "TyFreeVar z = vsubst_ty (TyFreeVar x) i \<tau>2" and "x \<noteq> z" by auto
  thus ?case by (cases \<tau>2, auto)
next
  case (TyVar y i \<tau>2) thus ?case proof (cases "y=i")
    case False
    hence "TyVar y = vsubst_ty (TyFreeVar x) i \<tau>2" using TyVar by auto
    thus ?thesis by (cases \<tau>2, auto)
  next
    case True
    have "\<tau>2 \<noteq> TyFreeVar x" using prems by auto
    moreover have "TyFreeVar x = vsubst_ty (TyFreeVar x) i \<tau>2" using TyVar True by auto
    ultimately show ?thesis proof (cases \<tau>2, auto)
      fix n
      assume "TyFreeVar x = (if n = i then TyFreeVar x else TyVar n)"
      thus "y=n" using True by (cases "n=i", auto)
    qed
  qed
next
  case (TyArrow \<tau>a \<tau>r i \<tau>2)
  hence "vsubst_ty (TyFreeVar x) i \<tau>a \<rightarrow> vsubst_ty (TyFreeVar x) i \<tau>r =
         vsubst_ty (TyFreeVar x) i \<tau>2" by auto
  thus ?case using prems proof (cases \<tau>2, auto)
    fix n
    assume "vsubst_ty (TyFreeVar x) i \<tau>a \<rightarrow> vsubst_ty (TyFreeVar x) i \<tau>r =
           (if n = i then TyFreeVar x else TyVar n)"
    thus False by (cases "n=i", auto)
  qed
next
  case (TyRef \<tau>ins i \<tau>2)
  hence "TyRef (vsubst_ty (TyFreeVar x) i \<tau>ins) = vsubst_ty (TyFreeVar x) i \<tau>2" by auto
  thus ?case using prems proof (cases \<tau>2, auto)
    fix n
    assume "TyRef (vsubst_ty (TyFreeVar x) i \<tau>ins) = (if n = i then TyFreeVar x else TyVar n)"
    thus False by (cases "n=i", auto)
  qed
next
  case (TyForall \<tau>ins i \<tau>2)
  hence "\<forall>.(vsubst_ty (TyFreeVar x) (Suc i) \<tau>ins) = vsubst_ty (TyFreeVar x) i \<tau>2" by auto
  thus ?case using prems proof (cases \<tau>2, auto)
    fix n
    assume "\<forall>. (vsubst_ty (TyFreeVar x) (Suc i) \<tau>ins) = (if n = i then TyFreeVar x else TyVar n)"
    thus False by (cases "n=i", auto)
  qed
qed

consts parametrize_ty :: "name \<Rightarrow> nat \<Rightarrow> ty \<Rightarrow> ty"
primrec
  "parametrize_ty a i TyUnit = TyUnit"
  "parametrize_ty a i (TyBase n) = TyBase n"
  "parametrize_ty a i (TyFreeVar b) = (if a=b then TyVar i else TyFreeVar b)"
  "parametrize_ty a i (TyVar n) = TyVar n"
  "parametrize_ty a i (\<tau>1\<rightarrow>\<tau>2) = (parametrize_ty a i \<tau>1) \<rightarrow> (parametrize_ty a i \<tau>2)"
  "parametrize_ty a i (\<forall>.\<tau>) = \<forall>.(parametrize_ty a (Suc i) \<tau>)"
  "parametrize_ty a i (TyRef \<tau>) = TyRef (parametrize_ty a i \<tau>)"

lemma parametrize_ty_sanity:
  assumes "i \<notin> set (free_db_ty \<tau>)"
  shows "\<not> a free in type parametrize_ty a i \<tau> \<and>
         \<tau> = vsubst_ty (TyFreeVar a) i (parametrize_ty a i \<tau>)"
proof -
  { fix n :: nat
    assume "size \<tau> = n"
      and "i \<notin> set (free_db_ty \<tau>)"
    hence "\<not> a free in type parametrize_ty a i \<tau> \<and>
           \<tau> = vsubst_ty (TyFreeVar a) i (parametrize_ty a i \<tau>)"
    proof (induct n arbitrary: i \<tau> rule: nat_less_induct)
      case (1 n i \<tau>) thus ?case proof (cases \<tau>)
        case (TyArrow \<tau>1 \<tau>2)
        { obtain n1 where "n1 = size \<tau>1" by auto
          moreover
          hence "n1 < n" using prems by auto
          moreover
          have "i \<notin> set (free_db_ty \<tau>1)" using prems by auto
          ultimately
          have "\<not> a free in type parametrize_ty a i \<tau>1 \<and>
                \<tau>1 = vsubst_ty (TyFreeVar a) i (parametrize_ty a i \<tau>1)" using "1.hyps" by auto
        }
        moreover
        { obtain n2 where "n2 = size \<tau>2" by auto
          moreover
          hence "n2 < n" using prems by auto
          moreover
          have "i \<notin> set (free_db_ty \<tau>2)" using prems by auto
          ultimately
          have "\<not> a free in type parametrize_ty a i \<tau>2 \<and>
                \<tau>2 = vsubst_ty (TyFreeVar a) i (parametrize_ty a i \<tau>2)" using "1.hyps" by auto
        }
        ultimately
        show ?thesis using TyArrow by auto
      next
        case (TyForall \<tau>ins)
        obtain nins where "nins = size \<tau>ins" by auto
        moreover
        hence "nins < n" using prems by auto
        moreover
        have "Suc i \<notin> set (free_db_ty \<tau>ins)" using prems by (auto simp add: shift_many_sanity)
        ultimately
        have "\<not> a free in type parametrize_ty a (Suc i) \<tau>ins \<and>
              \<tau>ins = vsubst_ty (TyFreeVar a) (Suc i) (parametrize_ty a (Suc i) \<tau>ins)"
          using "1.hyps" by auto
        thus ?thesis using TyForall by auto
      qed auto
    qed 
  }
  thus ?thesis using prems by auto
qed

lemma parametrize_ty_aux:
  assumes "i \<notin> set (free_db_ty \<tau>)"
  shows "\<exists>\<tau>'. \<not> x free in type \<tau>' \<and> \<tau> = vsubst_ty (TyFreeVar x) i \<tau>'"
using prems parametrize_ty_sanity by auto

typedef cty = "{\<tau> | \<tau>. set (free_db_ty \<tau>) = {}}"
   by (auto simp add: exI[of _ TyUnit])

consts not_free_in_func :: "name \<Rightarrow> (cty \<Rightarrow> ty) \<Rightarrow> bool"
defs not_free_in_func[simp]:
  "not_free_in_func a f \<equiv> (\<forall>\<tau>. \<not> a free in type (Rep_cty \<tau>) \<longrightarrow> \<not> a free in type f \<tau>)"

consts SubstFunc :: "(cty \<Rightarrow> ty) set"
syntax
  "subst_func" :: "(cty \<Rightarrow> ty) \<Rightarrow> bool"
translations
  "subst_func f" \<rightleftharpoons> "f \<in> SubstFunc"

inductive SubstFunc
  intros
  subst_func_id[intro!]      : "subst_func (%t. Rep_cty t)"
  subst_func_unit[intro!]    : "subst_func (%t. TyUnit)"
  subst_func_base[intro!]    : "subst_func (%t. TyBase n)"
  subst_func_freevar[intro!] : "subst_func (%t. TyFreeVar a)"
  subst_func_var[intro!]     : "subst_func (%t. TyVar k)"
  subst_func_arrow[intro!]   : "\<lbrakk> subst_func f; subst_func g \<rbrakk> \<Longrightarrow> subst_func (%t. f t \<rightarrow> g t)"
  subst_func_forall[intro!]  : "\<lbrakk> subst_func f \<rbrakk> \<Longrightarrow> subst_func (%t. \<forall>.f t)"
  subst_func_ref[intro!]     : "\<lbrakk> subst_func f \<rbrakk> \<Longrightarrow> subst_func (%t. TyRef (f t))"

lemma subst_func_const:
  shows "subst_func (%ct. \<tau>)"
by (induct \<tau>, auto)

lemma subst_func_subst:
  assumes "subst_func g"
  and "subst_func f"
  shows "subst_func (%t. vsubst_ty (f t) i (g t))"
using prems proof (induct g arbitrary: i f, auto)
  fix i f
  { fix t
    have "set (free_db_ty (Rep_cty t)) = {}" using cty_def Rep_cty by auto
    hence "i \<notin> set (free_db_ty (Rep_cty t))" by auto
    hence "vsubst_ty (f t) i (Rep_cty t) = Rep_cty t" by (rule not_in_free_db_same_subst_ty)
  }
  thus "subst_func (%t. vsubst_ty (f t) i (Rep_cty t))" by auto
next
  fix k i f
  assume "subst_func f"
  show "subst_func (%t. if k = i then f t else TyVar k)" by (cases "k = i", auto)
qed

declare Abs_cty_inverse[simp]
declare Rep_cty_inverse[simp]

lemma free_var_ty_sanity:
  shows "\<not> (\<forall>\<tau>. a free in type (Rep_cty \<tau>))"
proof
  assume "\<forall>\<tau>. a free in type (Rep_cty \<tau>)"
  hence "a free in type (Rep_cty (Abs_cty TyUnit))" by blast
  moreover
  have "\<not> a free in type TyUnit" by auto
  moreover
  have "set (free_db_ty TyUnit) = {}" by auto
  hence "TyUnit \<in> cty" by (auto simp add: cty_def)
  hence "Rep_cty (Abs_cty TyUnit) = TyUnit" by auto
  ultimately
  show False by auto
qed

lemma subst_func_equal_ty:
  assumes "subst_func f"
  and "subst_func g"
  and "not_free_in_func a f"
  and "not_free_in_func a g"
  and "f (Abs_cty (TyFreeVar a)) = g (Abs_cty (TyFreeVar a))"
  shows "f \<tau> = g \<tau>"
proof -
  have "set (free_db_ty (TyFreeVar a)) = {}" by auto
  hence "TyFreeVar a \<in> cty" by (auto simp add: cty_def)
  hence "Rep_cty (Abs_cty (TyFreeVar a)) = TyFreeVar a" by auto
  from prems this show ?thesis proof (induct f arbitrary: g \<tau>)
    case (subst_func_id g \<tau>) thus ?case proof (cases g)
      case (subst_func_freevar a')
      have "a \<noteq> a'" using prems not_free_in_func[of a g] by (auto simp add: free_var_ty_sanity)
      show ?thesis using prems not_free_in_func by (auto simp add: free_var_ty_sanity)
    qed auto
  next
    case (subst_func_arrow f1 f2 g \<tau>)
    have "subst_func g" .
    thus ?case using subst_func_arrow.prems proof (cases g, auto)
      fix g1 g2
      assume "\<forall>\<tau>. a \<notin> set (free_ty (Rep_cty \<tau>)) \<longrightarrow>
               a \<notin> set (free_ty (f1 \<tau>)) \<and> a \<notin> set (free_ty (f2 \<tau>))"
        and "\<forall>\<tau>. a \<notin> set (free_ty (Rep_cty \<tau>)) \<longrightarrow>
        a \<notin> set (free_ty (g1 \<tau>)) \<and> a \<notin> set (free_ty (g2 \<tau>))"
        and Heq: "f1 (Abs_cty (TyFreeVar a)) = g1 (Abs_cty (TyFreeVar a))"
        and Hsf: "subst_func g1"
        and Hx:  "Rep_cty (Abs_cty (TyFreeVar a)) = TyFreeVar a"
      hence "not_free_in_func a f1" and "not_free_in_func a g1" by auto
      thus "f1 \<tau> = g1 \<tau>" using subst_func_arrow.hyps Hsf Heq Hx by blast
    next
      fix g1 g2
      assume "\<forall>\<tau>. a \<notin> set (free_ty (Rep_cty \<tau>)) \<longrightarrow>
              a \<notin> set (free_ty (f1 \<tau>)) \<and> a \<notin> set (free_ty (f2 \<tau>))"
        and "\<forall>\<tau>. a \<notin> set (free_ty (Rep_cty \<tau>)) \<longrightarrow>
             a \<notin> set (free_ty (g1 \<tau>)) \<and> a \<notin> set (free_ty (g2 \<tau>))"
          and Heq: "f2 (Abs_cty (TyFreeVar a)) = g2 (Abs_cty (TyFreeVar a))"
          and Hsf: "subst_func g2"
          and Hx:  "Rep_cty (Abs_cty (TyFreeVar a)) = TyFreeVar a"
        hence "not_free_in_func a f2" and "not_free_in_func a g2" by auto
      thus "f2 \<tau> = g2 \<tau>" using subst_func_arrow.hyps Hsf Heq Hx by blast
    qed
  next
    case (subst_func_forall fins g \<tau>)
    have "subst_func g" .
    thus ?case using subst_func_forall.prems proof (cases g, auto)
      fix gins
      assume "\<forall>\<tau>. a \<notin> set (free_ty (Rep_cty \<tau>)) \<longrightarrow> a \<notin> set (free_ty (fins \<tau>))"
        and "\<forall>\<tau>. a \<notin> set (free_ty (Rep_cty \<tau>)) \<longrightarrow> a \<notin> set (free_ty (gins \<tau>))"
        and Heq: "fins (Abs_cty (TyFreeVar a)) = gins (Abs_cty (TyFreeVar a))"
        and Hsf: "subst_func gins"
        and Hx:  "Rep_cty (Abs_cty (TyFreeVar a)) = TyFreeVar a"
      hence "not_free_in_func a fins" and "not_free_in_func a gins" by auto
      thus "fins \<tau> = gins \<tau>" using subst_func_forall.hyps Hsf Heq Hx by blast
    qed
  next
    case (subst_func_ref fr g \<tau>)
    have "subst_func g" .
    thus ?case using subst_func_ref.prems proof (cases g, auto)
      fix gr
      assume "\<forall>\<tau>. a \<notin> set (free_ty (Rep_cty \<tau>)) \<longrightarrow> a \<notin> set (free_ty (fr \<tau>))"
        and "\<forall>\<tau>. a \<notin> set (free_ty (Rep_cty \<tau>)) \<longrightarrow> a \<notin> set (free_ty (gr \<tau>))"
        and Heq: "fr (Abs_cty (TyFreeVar a)) = gr (Abs_cty (TyFreeVar a))"
        and Hsf: "subst_func gr"
        and Hx:  "Rep_cty (Abs_cty (TyFreeVar a)) = TyFreeVar a"
      hence "not_free_in_func a fr" and "not_free_in_func a gr" by auto
      thus "fr \<tau> = gr \<tau>" using subst_func_ref.hyps Hsf Heq Hx by blast
    qed
  next
    case (subst_func_unit g \<tau>) thus ?case by (cases g, auto)
  next
    case (subst_func_base n g t) thus ?case by (cases g, auto)
  next
    case (subst_func_freevar b g \<tau>) thus ?case by (cases g, auto simp add: free_var_ty_sanity)
  next
    case (subst_func_var k g \<tau>) thus ?case by (cases g, auto)
  qed
qed

lemma vsubst_equal_sanity:
  assumes "vsubst_ty (vsubst_ty (TyFreeVar x) i \<tau>a) 0 (vsubst_ty (TyFreeVar x) (Suc i) \<tau>d) =
           vsubst_ty (TyFreeVar x) i \<tau>"
  and "\<not> x free in type \<tau>"
  and "\<not> x free in type \<tau>a"
  and "\<not> x free in type \<tau>d"
  and "set (free_db_ty \<tau>) \<subseteq> {i}"
  and "set (free_db_ty \<tau>a) \<subseteq> {i}"
  and "set (free_db_ty \<tau>d) \<subseteq> {0, Suc i}"
  and "set (free_db_ty \<tau>') = {}"
  shows "vsubst_ty (vsubst_ty \<tau>' i \<tau>a) 0 (vsubst_ty \<tau>' (Suc i) \<tau>d) = vsubst_ty \<tau>' i \<tau>"
proof -
  let ?f = "%ct. vsubst_ty (vsubst_ty (Rep_cty ct) i \<tau>a) 0 (vsubst_ty (Rep_cty ct) (Suc i) \<tau>d)"
  let ?g = "%ct. vsubst_ty (Rep_cty ct) i \<tau>"
  let ?\<tau>'' = "Rep_cty (Abs_cty \<tau>')"
  have "subst_func (%ct. vsubst_ty (Rep_cty ct) (Suc i) \<tau>d)"
    using subst_func_subst[of "%ct. \<tau>d"] by (auto simp add: subst_func_const)
  moreover
  have "subst_func (%ct. vsubst_ty (Rep_cty ct) i \<tau>a)"
    using subst_func_subst[of "%ct. \<tau>a"] by (auto simp add: subst_func_const)
  ultimately
  have "subst_func ?f" using subst_func_subst by auto
  moreover
  have "subst_func ?g" using subst_func_subst[of "%ct. \<tau>"] by (auto simp add: subst_func_const)
  moreover
  have "not_free_in_func x ?f" using prems by (auto simp add: not_free_subst_ty)
  moreover
  have "not_free_in_func x ?g" using prems by (auto simp add: not_free_subst_ty)
  moreover
  have "set (free_db_ty (TyFreeVar x)) = {}" by auto
  hence "TyFreeVar x \<in> cty" by (auto simp add: cty_def)
  hence "Rep_cty (Abs_cty (TyFreeVar x)) = TyFreeVar x" by auto
  hence "?f (Abs_cty (TyFreeVar x)) = ?g (Abs_cty (TyFreeVar x))" by auto
  ultimately
  have "vsubst_ty (vsubst_ty ?\<tau>'' i \<tau>a) 0 (vsubst_ty ?\<tau>'' (Suc i) \<tau>d) = vsubst_ty ?\<tau>'' i \<tau>"
    by (rule subst_func_equal_ty)
  moreover
  have "\<tau>' \<in> cty" using prems by (auto simp add: cty_def)
  hence "?\<tau>'' = \<tau>'" by auto
  ultimately
  show ?thesis by auto
qed

end
